import pyrosetta

try:
    from rdkit_to_params import Params, neutralise
except ModuleNotFoundError:
    pass
except ImportError:
    pass
from typing import (Optional, Union, Dict, List)
from ..common_ops.downloads import download_pdb
from .load import parameterised_pose_from_file, parameterised_pose_from_pdbblock
from io import IOBase  # just for an isinstance fh
import warnings

pr_rs = pyrosetta.rosetta.core.select.residue_selector


def chain_letter_to_number(letter, pose):
    warnings.warn('this is a shitty interim way of going from letter to number...' + \
                  ' I am sure there"s a method somewhere that does this.')
    pdb_info = pose.pdb_info()
    for i in range(1, pose.num_chains() + 1):
        b = pose.chain_begin(i)
        if pdb_info.chain(b) == letter:
            return i
    else:
        raise ValueError


class LigandNicker:
    """
    Given a pdb_file (regular initialisation) or code (``.from_pdbcode`` classmethod)
    and a list of 3-letter codes of wanted residues (``wanted_ligands`` argument),
    it loads it as Pyrosetta Pose (``donor_pose``), ready for ``migrate`` to loads the acceptor_pose and nick
    the residues that are wanted.

    If ``force_parameterisation`` is on it or the residue is novel it parameterises it.
    """

    def __init__(self,
                 pdb_filename: Optional[str] = None,
                 pdb_filehandle: Optional[IOBase] = None,
                 pdb_block: Optional[str] = None,
                 pose: Optional[pyrosetta.Pose] = None,
                 chain: str = 'A',
                 wanted_ligands: List[str] = (),
                 force_parameterisation: bool = False,
                 neutralise_params: bool = True,
                 save_params: bool = True,
                 overriding_params=()):
        """
        Initialisation loads the donor. ``migrate`` loads the acceptor.

        Unfortunately, I originally wrote it to use pdb_filename only.
        I should have written the init to accept a pose and
        a class method the filename. And now some chucks of out there in the wild use pdb_filename.
        So I could not switch it to something generic. Now it accepts four possible alternative choices:
        pdb_filename, pdb_filehandle, pdb_block, pose.


        :param pdb_filename:
        :param pdb_filehandle:
        :param pdb_block:
        :param pose:
        :param chain:
        :param wanted_ligands:
        :param force_parameterisation:
        :param neutralise_params: pH 7 protonation
        :param save_params:
        :param overriding_params: overide paramaterisation and use provide params
        """

        self.donor_chain = chain
        assert len(wanted_ligands), 'No wanted_ligands specified'
        self.wanted_ligands = wanted_ligands
        self.extra_params = []
        self.acceptor_pose = None
        # ---- rubbish overloading  ------
        self.pdb_filename = pdb_filename
        if pose:
            assert isinstance(pose, pyrosetta.Pose) and pose.total_residue()
            self.pdb_filename = ''
            self.donor_pose = pose
        elif pdb_filename:
            assert isinstance(pdb_filename, str)
            self.donor_pose = parameterised_pose_from_file(pdb_filename=pdb_filename,
                                                           force_parameterisation=force_parameterisation,
                                                           neutralise_params=neutralise_params,
                                                           save_params=save_params,
                                                           overriding_params=overriding_params)
        elif pdb_filehandle or pdb_block:
            if pdb_filehandle:
                assert isinstance(pdb_filehandle, IOBase)
                pdb_block = pdb_filehandle.read()
            else:
                pass
            self.donor_pose = parameterised_pose_from_pdbblock(pdbblock=pdb_block,
                                                               force_parameterisation=force_parameterisation,
                                                               neutralise_params=neutralise_params,
                                                               save_params=save_params,
                                                               overriding_params=overriding_params
                                                               )
        else:
            raise TypeError(f'Please provide with one of pdb_filename, pdb_filehandle, pdb_block, pose')

    @classmethod
    def from_pdbcode(cls, pdb_code: str, chain: str, *args, **kvargs):
        pdb_filename = download_pdb(pdb_code)
        return cls(pdb_filename=pdb_filename, chain=chain, *args, **kvargs)

    def migrate(self, acceptor_pose: pyrosetta.Pose, acceptor_chain: str = 'A',
                constrained: bool = True, relaxed: bool = True,
                relax_radius: int = 20, relax_cycles: int = 3):
        """
        The acceptor pose is the non-empty pose.

        This method aligns the sequences of the acceptor and donor pose.
        It finds the mapping of the neighbourhood of the wanted residues of the donor_pose
        It superimposes the poses by those residues.
        It adds the residues that need nicking.
        It adds constraints (optionally) based on the hydrogen bonding of the residues around the wanted residues.
        onto the ``.acceptor_pose``. To check:

        >>> print( len(self.acceptor_pose.constraint_set().get_all_constraints())  )

        It then optionally relaxes the neighbourhood.

        :param acceptor_pose:
        :param acceptor_chain:
        :param constrained:
        :param relaxed:
        :param relax_radius:
        :param relax_cycles:
        :return:
        """
        # store
        self.acceptor_pose = acceptor_pose
        self.acceptor_chain = acceptor_chain
        # get wanted residues
        wanted_selector = self.get_wanted_selector()
        donor_neighbors = self.get_surrounding_residue(self.donor_pose,
                                                       self.donor_chain,
                                                       wanted_selector)
        # get the neighbouring residues & superpose w/ them
        dex = self.get_mapping_between_poses(donor_neighbors)
        mapping = self.make_atomID_map(dex, self.donor_pose, self.acceptor_pose)
        pyrosetta.rosetta.core.scoring.superimpose_pose(self.donor_pose,
                                                        self.acceptor_pose,
                                                        mapping)
        wanted_vector = wanted_selector.apply(self.donor_pose)
        # append the ligands. n nucleotides have issues with the grafting command.
        # hence this way.
        added_indices = []
        for r in pr_rs.ResidueVector(wanted_vector):
            r2 = self.acceptor_pose.total_residue() + 1
            dex[r] = r2
            added_indices.append(r2)
            # append_subpose_to_pose is cool with novel residues.
            pyrosetta.rosetta.core.pose.append_subpose_to_pose(self.acceptor_pose,
                                                               self.donor_pose,
                                                               r, r, True)
        # make selector of newly added residues
        self.added_selector = pr_rs.ResidueIndexSelector()
        for r in added_indices:
            self.added_selector.append_index(r)
        # constrain
        if constrained:
            self.constrain_migrated(wanted_vector, dex)
        # relax
        if relaxed:
            self.relax_migrated(distance=relax_radius, cycles=relax_cycles)

    # ---- migrate dependent methods

    def get_wanted_selector(self):
        # select the ligands that are wanted.
        # set_residue_names does not like custom residue types.
        wanted_ligands = {c.rjust(3) for c in self.wanted_ligands}
        # ResidueNameSelector seems to not like custom residues if they arent in the pose
        try:
            resn_sele = pr_rs.ResidueNameSelector()
            resn_sele.set_residue_name3(','.join(wanted_ligands))
            resn_sele.apply(self.donor_pose)
        except RuntimeError as err:
            raise ValueError(f'Residue not in the pose. {err}')
        # end of shitty hack.
        # filter for those within 3A of chain of interest
        # define extended neighbours
        chain_sele = pyrosetta.rosetta.core.select.residue_selector.ChainSelector(self.donor_chain)
        cc_sele = pyrosetta.rosetta.core.select.residue_selector.CloseContactResidueSelector()
        cc_sele.central_residue_group_selector(chain_sele)
        cc_sele.threshold(3)
        # intersection:
        and_sele = pyrosetta.rosetta.core.select.residue_selector.AndResidueSelector(cc_sele, resn_sele)
        return and_sele

    def get_surrounding_residue(self, pose, chain_filter, wanted_selector):
        cc_sele = pyrosetta.rosetta.core.select.residue_selector.CloseContactResidueSelector()
        cc_sele.central_residue_group_selector(wanted_selector)
        cc_sele.threshold(3)
        prop_sele = pr_rs.ResiduePropertySelector()
        prop_sele.add_property(pyrosetta.rosetta.core.chemical.ResidueProperty.ALPHA_AA)
        neigh_sele = pr_rs.AndResidueSelector(prop_sele, cc_sele)
        neigh_vector = neigh_sele.apply(pose)
        pdb_info = pose.pdb_info()
        rs = pr_rs.ResidueVector(neigh_vector)
        neigh_resis = [r for r in rs if pdb_info.chain(r) == chain_filter]
        return neigh_resis  # these are the pose indices

    def _make_map(self, al, pose_offset=0):
        """
        pose index (fortran-style) to MSA index (C++-style)

        Returns a dictionary of key = offset + ungapped index (zero-indexed) to msa index  (zero-indexed)

        :param al:
        :param pose_offset:
        :return:
        """
        gap_map = [i for i, r in enumerate(al) if r != '-']
        return {pose_offset + ungap_i: gap_i for ungap_i, gap_i in enumerate(gap_map)}

    def get_mapping_between_poses(self, index_list: List[int]) -> Dict[int, int]:
        """
        Given a list of indices for one pose, return their aligned equivalents in the second.
        Modded to be donor_pose --> acceptor_pose
        """
        from Bio import pairwise2
        acc_ch_idx = chain_letter_to_number(self.acceptor_chain, self.acceptor_pose)
        don_ch_idx = chain_letter_to_number(self.donor_chain, self.donor_pose)
        acc_seq = self.acceptor_pose.chain_sequence(acc_ch_idx)
        don_seq = self.donor_pose.chain_sequence(don_ch_idx)
        alignments = pairwise2.align.globalxs(acc_seq,
                                              don_seq,
                                              -1,  # open
                                              -0.1  # extend
                                              )
        alignments = dict(zip(['target', 'template', 'score', 'begin', 'end'], alignments[0]))

        donor_pose_to_msa_mapping = self._make_map(alignments['template'],
                                                   self.donor_pose.chain_begin(don_ch_idx))
        acceptor_pose_to_msa_mapping = self._make_map(alignments['target'],
                                                      self.acceptor_pose.chain_begin(acc_ch_idx))
        msa_to_acceptor_pose_mapping = dict(zip(acceptor_pose_to_msa_mapping.values(),
                                                acceptor_pose_to_msa_mapping.keys()))
        mapping = {}
        for r in index_list:
            msa_i = donor_pose_to_msa_mapping[r]
            trans = msa_to_acceptor_pose_mapping[msa_i]
            #             print(r, self.donor_pose.residue(r).name1(),
            #                   msa_i, alignments['template'][msa_i],
            #                   trans, self.acceptor_pose.residue(trans).name1() )
            mapping[r] = trans
        return mapping

    def make_atomID_map(self, dex, query_pose, target_pose):
        mapping = pyrosetta.rosetta.std.map_core_id_AtomID_core_id_AtomID()
        get_id = lambda pose, r, atomname: pyrosetta.AtomID(atomno_in=pose.residue(r).atom_index(atomname), rsd_in=r)
        for q, t in dex.items():
            for atomname in ('CA',):
                mapping[get_id(query_pose, q, atomname)] = get_id(target_pose, t, atomname)
        return mapping

    def constrain_migrated(self, wanted_vector, dex):
        hbond_set = self.donor_pose.get_hbonds()
        wanted_indices = pr_rs.ResidueVector(wanted_vector)
        for hbond in hbond_set.hbonds():
            if hbond.don_res() in wanted_indices or hbond.acc_res() in wanted_indices:
                con = self.make_constraint_foreign_hbond(hbond, dex)
                if con:
                    self.acceptor_pose.add_constraint(con)

    def make_constraint_foreign_hbond(self, hbond, dex: Dict[int, int]) \
            -> Union[None, pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint]:
        # acc = hydrogen acceptor residue w/Acceptor atom
        # don = donor residue w/ Bounded and Hydrogen atoms.
        # ------------- sort out donor -------------
        don_res = hbond.don_res()
        don = self.donor_pose.residue(don_res)
        if don_res not in dex:
            warnings.warn(
                f'The H-bond–donor residue in the donor pose has residue {don_res} with no equivalent - why was it absent in CC selector?')
            return
        trans_don_res = dex[don_res]  # acceptor_pose
        trans_don = self.acceptor_pose.residue(trans_don_res)
        don_name3 = don.name3()
        trans_don_name3 = trans_don.name3()
        if not hbond.don_hatm_is_backbone() and don_name3 != trans_don_name3:
            return  # skip as it is a sidechain Hbond of a changed residue.
        elif not hbond.don_hatm_is_backbone() and don_name3 == 'HIS':
            return  # avoiding HID vs HIE pain
        don_atomname = don.atom_name(hbond.don_hatm())
        donor_atom = pyrosetta.AtomID(atomno_in=trans_don.atom_index(don_atomname),
                                      rsd_in=trans_don_res)
        # ------------- sort out acceptor -------------
        acc_res = hbond.acc_res()
        acc = self.donor_pose.residue(acc_res)
        trans_acc_res = dex[acc_res]  # acceptor_pose
        acc_name3 = acc.name3()
        trans_acc = self.acceptor_pose.residue(trans_acc_res)
        trans_acc_name3 = trans_acc.name3()
        if not hbond.acc_atm_is_backbone() and acc_name3 != trans_acc_name3:
            return  # skip as it is a sidechain Hbond of a changed residue.
        elif not hbond.acc_atm_is_backbone() and acc_name3 == 'HIS':
            return  # avoiding HID vs HIE pain
        acc_atomname = acc.atom_name(hbond.acc_atm())
        acceptor_atom = pyrosetta.AtomID(atomno_in=trans_acc.atom_index(acc_atomname),
                                         rsd_in=trans_acc_res)
        # ------------- make constraint -------------
        HarmonicFunc = pyrosetta.rosetta.core.scoring.func.HarmonicFunc
        AtomPairConstraint = pyrosetta.rosetta.core.scoring.constraints.AtomPairConstraint
        d = hbond.get_HAdist(self.donor_pose)
        return AtomPairConstraint(donor_atom,
                                  acceptor_atom,
                                  HarmonicFunc(x0_in=d, sd_in=0.2))

    def relax_migrated(self, distance: int = 20, cycles: int = 3, atom_pair_weight: int = 5):
        scorefxn = pyrosetta.get_fa_scorefxn()
        atom_pair = pyrosetta.rosetta.core.scoring.ScoreType.atom_pair_constraint
        scorefxn.set_weight(atom_pair, atom_pair_weight)
        neigh_sele = pr_rs.NeighborhoodResidueSelector(self.added_selector, distance=distance,
                                                       include_focus_in_subset=True)
        movemap = pyrosetta.MoveMap()
        n = neigh_sele.apply(self.acceptor_pose)
        movemap.set_bb(allow_bb=n)
        movemap.set_chi(allow_chi=n)
        ft = self.acceptor_pose.fold_tree()
        for r in pr_rs.ResidueVector(self.added_selector.apply(self.acceptor_pose)):
            j = ft.get_jump_that_builds_residue(r)
            movemap.set_jump(j, True)
        relax = pyrosetta.rosetta.protocols.relax.FastRelax(scorefxn, cycles)
        relax.set_movemap_disables_packing_of_fixed_chi_positions(True)
        relax.set_movemap(movemap)
        relax.apply(self.acceptor_pose)



from typing import (Optional, Tuple, Iterable, Dict)

import pyrosetta

__all__ = ['get_alignment', 'write_grishin', 'thread', 'rangify', 'steal_ligands',
           'get_nonprotein_pose', 'oligomer_thread', 'make_fragment_sets']


def get_alignment(target: str, template: str) -> Dict[str, str]:
    """
    Returns alignments using ``pairwise2.align.globalxs``
    """
    from Bio import pairwise2
    alignments = pairwise2.align.globalxs(target,
                                          template,
                                          -1,  # open
                                          -0.1  # extend
                                          )
    return dict(zip(['target', 'template', 'score', 'begin', 'end'], alignments[0]))


def write_grishin(target_name, target_sequence, template_name, template_sequence, outfile):
    with open(outfile, 'w') as w:
        w.write(f'## {target_name} {template_name}\n')
        w.write(f'#\n')
        w.write('scores_from_program: 0\n')
        w.write(f'0 {target_sequence}\n')
        w.write(f'0 {template_sequence}\n')
        w.write('--\n')


def thread(target_sequence: str,
           template_pose: pyrosetta.Pose,
           target_pose: Optional[pyrosetta.Pose] = None,
           target_name: str = 'target',
           template_name: str = 'template',
           fragment_sets: Optional[pyrosetta.rosetta.utility.vector1_std_shared_ptr_core_fragment_FragSet_t] = None,
           align: Optional[pyrosetta.rosetta.core.sequence.SequenceAlignment] = None
           ) -> Tuple[pyrosetta.Pose,
                      pyrosetta.rosetta.protocols.comparative_modeling.ThreadingMover,
                      pyrosetta.rosetta.utility.vector1_bool]:
    """
    Given the target sequence and the optional blank target pose (in case there's some reason for it),
    thread the sequence against the template pose — which is assumed to be a single chain.
    Optionally using fragments from fragment_sets.
    The three outputs are the target_pose, the threader instance and a vector of the residues threaded.

    >>> print(threader.frag_libs()[1].nr_frames())
    
    >>> qt = threader.get_qt_mapping(threaded)
    >>> print(f'Template residue {21} is residue {q[21]} in target')

    :param target_sequence:
    :param template_pose:
    :param target_pose:
    :param target_name:
    :param template_name:
    :param fragment_sets:
    :param align:
    :return:
    """
    ## Make target_pose
    if target_pose is None:
        target_pose = pyrosetta.Pose()
        pyrosetta.rosetta.core.pose.make_pose_from_sequence(target_pose,
                                                            target_sequence,
                                                            'fa_standard')
    ## Unterminate and remove ligands
    original_template_pose = template_pose
    clean_template_pose = template_pose.clone()
    pyrosetta.rosetta.core.pose.remove_nonprotein_residues(clean_template_pose)
    ### find
    lowers = pyrosetta.rosetta.utility.vector1_std_pair_unsigned_long_protocols_sic_dock_Vec3_t()
    uppers = pyrosetta.rosetta.utility.vector1_std_pair_unsigned_long_protocols_sic_dock_Vec3_t()
    pyrosetta.rosetta.protocols.sic_dock.get_termini_from_pose(clean_template_pose, lowers, uppers)
    ### remove
    rm_upper = pyrosetta.rosetta.core.conformation.remove_upper_terminus_type_from_conformation_residue
    rm_lower = pyrosetta.rosetta.core.conformation.remove_lower_terminus_type_from_conformation_residue
    for upper, _ in uppers:
        rm_upper(clean_template_pose.conformation(), upper)
    for lower, _ in lowers:
        rm_lower(clean_template_pose.conformation(), lower)
    ## Align
    if align is None:
        alignment = get_alignment(target_sequence, clean_template_pose.sequence())
        aln_file = f'{target_name}.aln'
        write_grishin(target_name,
                      alignment['target'],
                      template_name,
                      alignment['template'],
                      aln_file)
        align = pyrosetta.rosetta.core.sequence.read_aln(format='grishin', filename=aln_file)[1]
    ## Thread
    # not pyrosetta.rosetta.protocols.comparative_modeling.PartialThreadingMover
    threader = pyrosetta.rosetta.protocols.comparative_modeling.ThreadingMover(align=align,
                                                                               template_pose=clean_template_pose)
    ### Input frags
    if fragment_sets is not None:
        threader.build_loops(True)
        threader.randomize_loop_coords(True)  # default
        threader.frag_libs(fragment_sets)
    ### Apply
    threader.apply(target_pose)
    ## Steal sidechains
    qt = threader.get_qt_mapping(target_pose)
    steal = pyrosetta.rosetta.protocols.comparative_modeling.StealSideChainsMover(clean_template_pose, qt)
    steal.apply(target_pose)
    ## Make vector of what was changed
    vector = pyrosetta.rosetta.utility.vector1_bool(target_pose.total_residue())
    mapping = qt.mapping()
    for r in range(1, target_pose.total_residue() + 1):
        if mapping[r] != 0:
            vector[r] = 1
    return target_pose, threader, vector


vector1_FragSet_t = pyrosetta.rosetta.utility.vector1_std_shared_ptr_core_fragment_FragSet_t


def make_fragment_sets(*poses: Iterable[pyrosetta.Pose], lengths: Iterable[int] = (3,)) -> vector1_FragSet_t:
    fragsets = vector1_FragSet_t(len(lengths))
    for i, l in enumerate(lengths):
        fragsets[i+1] = pyrosetta.rosetta.core.fragment.ConstantLengthFragSet(l)
        for pose in poses:
            pyrosetta.rosetta.core.fragment.steal_constant_length_frag_set_from_pose(pose, fragsets[i+1])
    return fragsets

def steal_ligands(donor_pose, acceptor_pose) -> None:
    """
    Steals non-Protein residues from donor_pose and adds them to acceptor_pose

    Do not use with nucleic acid polymers.

    :param donor_pose:
    :param acceptor_pose:
    :return:
    """
    PROTEIN = pyrosetta.rosetta.core.chemical.ResidueProperty.PROTEIN
    prot_sele = pyrosetta.rosetta.core.select.residue_selector.ResiduePropertySelector(PROTEIN)
    not_sele = pyrosetta.rosetta.core.select.residue_selector.NotResidueSelector(prot_sele)
    rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(not_sele.apply(donor_pose))
    # if it were DNA...
    # for from_res, to_res in rangify(rv):
    #     pyrosetta.rosetta.core.pose.append_subpose_to_pose(acceptor_pose, donor_pose, from_res, to_res, True)
    for res in rv:
        pyrosetta.rosetta.core.pose.append_subpose_to_pose(acceptor_pose, donor_pose, res, res, True)


def rangify(values):
    """
    Given a list of integers, returns a list of tuples of ranges (interger pairs).

    :param values:
    :return:
    """
    previous = None
    start = None
    ranges = []
    for r in values:
        if previous is None:
            previous = r
            start = r
        elif r == previous + 1:
            pass
        else:  # r != previous + 1
            ranges.append((start, previous))
            start = r
        previous = r
    ranges.append((start, previous))
    return ranges


def get_nonprotein_pose(pose):
    """
    pyrosetta.rosetta.protocols.comparative_modeling.StealLigandMover requires some weird things.
    This does the same, makes a ligand only pose.

    :param pose:
    :return:
    """
    protein_prop = pyrosetta.rosetta.core.chemical.ResidueProperty.PROTEIN
    prot_sele = pyrosetta.rosetta.core.select.residue_selector.ResiduePropertySelector(protein_prop)
    np_sele = pyrosetta.rosetta.core.select.residue_selector.NotResidueSelector(prot_sele)
    rv = pyrosetta.rosetta.core.select.residue_selector.ResidueVector(np_sele.apply(pose))
    if len(rv) == 0:
        return None
    poses = []
    for begin, end in rangify(rv):
        poses.append(pyrosetta.rosetta.protocols.grafting.return_region(pose, begin, end))
    neo = poses[0]
    for pose in poses[1:]:
        neo.append_pose_by_jump(pose, neo.num_jump() + 1)
    return neo


def oligomer_thread(pose, sequence):
    faux = pose.clone()
    pyrosetta.rosetta.core.pose.remove_nonprotein_residues(faux)
    threaded = None
    threaders = []
    master_vector = pyrosetta.rosetta.utility.vector1_bool()
    for chain_pose in faux.split_by_chain():
        target_pose, threader, vector = thread(target_sequence=sequence, template_pose=chain_pose)
        threaders.append(threader)
        if threaded is None:
            threaded = target_pose
        else:
            threaded.append_pose_by_jump(target_pose, threaded.num_jump() + 1)
            master_vector.extend(vector)
    threaded.append_pose_by_jump(get_nonprotein_pose(pose), threaded.num_jump() + 1)
    return threaded, threaders, master_vector



"""
Adds to the NGLView Widget the method ``.add_selector``

>>> import nglview as nv
>>> view = nv.view_rosetta(pose)
>>> view.add_selector(pose, selector)
>>> view
"""

from io import StringIO
from typing import (Optional)

import nglview
import pyrosetta

from .constraints import get_NGL_selection_from_AtomID
from .utils import get_pdbstr


def selector_to_ngl(self: nglview.widget.NGLWidget,
                    pose: pyrosetta.Pose,
                    selector: pyrosetta.rosetta.core.select.residue_selector.ResidueSelector) -> str:
    """
    Given a pose and a selector return the selection string for NGL.

    :param self:
    :param pose:
    :param selector:
    :return:
    """
    pdb_info = pose.pdb_info()
    # should probably `deal with pdb_info.segmentID(1).strip()`
    ResidueVector = pyrosetta.rosetta.core.select.residue_selector.ResidueVector
    selections = [f'{pdb_info.number(r)}:{pdb_info.chain(r)}' for r in ResidueVector(selector.apply(pose))]  # noqa
    selection = ' or '.join(selections)
    return selection


def add_selector(self: nglview.widget.NGLWidget,
                 pose: pyrosetta.Pose,
                 selector: pyrosetta.rosetta.core.select.residue_selector.ResidueSelector,
                 representation_name: str = 'hyperball',
                 color: str = 'grey',
                 **other):
    """
    Add a representation of type ``representation_name`` (def. 'hyperball') and center
    based upon the Pyrosetta residue selector

    :param self:
    :param pose:
    :param selector:
    :param representation_name:
    :param color:
    :param other:
    :return:
    """
    selection = self.selector_to_ngl(pose, selector)  # noqa --> Monkeypatched
    self.add_representation(representation_name,
                            colorValue=color,
                            selection=selection,
                            **other)
    self.center(selection)


def add_rosetta(self: nglview.widget.NGLWidget,
                pose: pyrosetta.Pose, color: Optional[str] = None):
    """
    The module method ``show_rosetta`` creates an NGLWidget
    This is a monkeypatched bound method.

    :param self:
    :param pose:
    :return: component
    """
    fh = StringIO(get_pdbstr(pose))
    c = self.add_component(fh, ext='pdb')
    if color:
        c.update_cartoon(color='color', smoothSheet=True)
    return c


def make_pose_comparison(self: nglview.widget.NGLWidget,
                         first_pose: pyrosetta.Pose,
                         second_pose: pyrosetta.Pose,
                         first_color: str = '#00B4C4',
                         second_color: str = '#F8766D'):
    """
    Adds two objects, the first colored by default in #00B4C4, which is turquoise,
     while the second #F8766D, which is salmon.
     The poses are assumed aligned.

    :param self:
    :param first_pose:
    :param second_pose:
    :param first_color:
    :param second_color:
    :return:
    """
    c0 = self.add_rosetta(first_pose)  # noqa --> Monkeypatched
    c0.update_cartoon(color=first_color, smoothSheet=True)
    c1 = self.add_rosetta(second_pose)  # noqa --> Monkeypatched
    c1.update_cartoon(color=second_color, smoothSheet=True)


def add_constraints(self: nglview.widget.NGLWidget,
                    pose: pyrosetta.Pose,
                    component=0,
                    color="skyblue"):
    atom_pairs = [[get_NGL_selection_from_AtomID(pose, con.atom1(), named=False),
                   get_NGL_selection_from_AtomID(pose, con.atom2(), named=False)
                   ]
                  for con in pose.constraint_set().get_all_constraints()
                  ]
    atom_pairs = [[a1, a2] for a1, a2 in atom_pairs if a1 != a2]
    component = getattr(self, f'component_{component}')
    component.add_representation("distance",
                                 atomPair=atom_pairs,
                                 color=color
                                 )
    return atom_pairs


# ======= Monkey patch ==========================================

nglview.widget.NGLWidget.selector_to_ngl = selector_to_ngl
nglview.widget.NGLWidget.add_selector = add_selector
nglview.widget.NGLWidget.add_rosetta = add_rosetta
nglview.widget.NGLWidget.add_constraints = add_constraints
nglview.widget.NGLWidget.make_pose_comparison = make_pose_comparison



from warnings import warn
from importlib import util
import sys

if sys.version_info.major != 3 or sys.version_info.minor < 6:
    print(sys.version_info)
    raise SystemError('Module written for Python 3.6+.')

# ---------- Non pip modules  ------------------------------------------------------------------------------------------

if not util.find_spec('pyrosetta'):
    warn('This 3.6+ script **requires** pyrosetta, which has to be downloaded from ' +
         'the Rosetta software site due to licencing.')

# ---------- Setup  ------------------------------------------------------------------------------------------

import os
this_directory = os.path.abspath(os.path.dirname(__file__))
try:
    with open(os.path.join(this_directory,'requirements.txt'), 'r') as fh:
        requirements = fh.read().strip().split()
    with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        __doc__ = f.read()
except Exception:  # weird file
    requirements = ['rdkit-to-params', 'xerox']
    __doc__ = ''


from setuptools import setup, find_packages

description = 'A variety of functions to make working with Pyrosetta easier.'

setup(
    name='pyrosetta_help',
    version='0.4.7',
    python_requires='>=3.7',
    packages=find_packages(),
    install_requires=requirements,
    url='https://github.com/matteoferla/pyrosetta_help',
    license='MIT',
    author='Matteo Ferla',
    author_email='matteo.ferla@gmail.com',
    classifiers=[ # https://pypi.org/classifiers/
        'Development Status :: 4 - Beta', # Development Status :: 5 - Production/Stable
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    description=description,
    long_description=__doc__,
    long_description_content_type='text/markdown'
)



from typing import (Optional)

import pyrosetta


class ResInfo:
    pdb_info = None  #: pdb_info: pyrosetta.rosetta.core.pose.PDBInfo

    def __init__(self, number: int, chain: str, icode: str = ' ', segmentID: str = '    '):
        self.number = number
        self.chain = chain
        self.icode = icode
        self.segmentID = segmentID

    @classmethod
    def get(cls, index: int):
        return cls(number=cls.pdb_info.number(index),
                   chain=cls.pdb_info.chain(index),
                   icode=cls.pdb_info.icode(index),
                   segmentID=cls.pdb_info.segmentID(index))

    def set(self,
            index: int,
            pdb_info: Optional[pyrosetta.rosetta.core.pose.PDBInfo] = None):
        if pdb_info is None:
            pdb_info = self.pdb_info
        pdb_info.set_resinfo(res=index,
                             chain_id=self.chain,
                             pdb_res=self.number,
                             ins_code=self.icode,
                             segmentID=self.segmentID)


class BlueCopier:
    def copy_pdb_info(self, original_pose: pyrosetta.Pose, final_pose: pyrosetta.Pose):
        # get original residue info
        ResInfo.pdb_info = original_pose.pdb_info()
        original = []
        previous = None
        for row in self:
            if row[0] == 0:
                # insertion code is a letter
                previous.icode
                raise NotImplementedError
            ri = ResInfo.get(row[0])
            original.append(ri)
            previous = ri
        # set info
        pdb_info = final_pose.pdb_info()
        for i, row in enumerate(original):
            row.set(i + 1, pdb_info)
        pdb_info().obsolete(False)



from .constraints import *
from .downloads import *
from .utils import *
from .minimise import *
from .faux_selectors import *
from .ss_changes import *
try:
    from .nglview import nglview
except ImportError:
    pass

